/*
  Copyright (c) 2015 DataLab, s.l. <http://www.datalab.es>
  This file is part of GlusterFS.

  This file is licensed to you under your choice of the GNU Lesser
  General Public License, version 3 or any later version (LGPLv3 or
  later), or the GNU General Public License, version 2 (GPLv2), in all
  cases as published by the Free Software Foundation.
*/

#ifndef __EC_CODE_SSE_H__
#define __EC_CODE_SSE_H__

#include "ec-code.h"

extern ec_code_gen_t ec_code_gen_sse;

#endif /* __EC_CODE_SSE_H__ */
